/*******************************************************************************
 * cell.cpp - the development part of the exploration
 * class representing the cell, cellarray and cellcollection objects used during 
 * development
 *
 * Copyright (C) 2006 Joacim Thomassen
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or any later version. 
 * (See accompanying file LICENSE.txt or copy at 
 * http://www.gnu.org/copyleft/gpl.html)
 */
#include <iostream>
#include "cell.hpp"
#include "rule.hpp"

/* FIXME: If the flexibility to add/remove celltypes will be more prominent
 * in future versions, it might be an idea to push all celltypes into a
 * vector and just assign static's to each element in the initial part of
 * the program (if possible). The vector size will give number_of_types. */
const int Cell::GROW_NORTH = 1;
const int Cell::GROW_SOUTH = 2;
const int Cell::GROW_EAST  = 3;
const int Cell::GROW_WEST  = 4;
const int Cell::TYPE_BLACK = 5;
const int Cell::TYPE_WHITE = 6;
const int Cell::DONT_CARE  = 7;
const int Cell::EMPTY      = 8;

const int Cell::TYPE_RED   = 9;
const int Cell::TYPE_BLUE  = 10;
const int Cell::TYPE_GREEN = 11;

// Use this static varible to constrain the number of types
// The number of visible/phenotypic types are hardcoded to 3 for now
const int Cell::NUMBER_OF_TYPES  = 11;


#define CELL_DEBUG(x)
//#define CELL_DEBUG(x) x

Cell::Cell():
   m_type(EMPTY)
{
  CELL_DEBUG(std::cout << "Creating cell of type " << m_type << std::endl);
}

Cell::Cell(int type)
   : m_type(type)
{
#if 0
   // REMOVED SINCE IT HIDES THE ERROR TO THE LOG
   if (m_type > NUMBER_OF_TYPES)
      m_type = EMPTY;
   
   if (m_type < 0) {
      m_type = EMPTY;
   }
#endif
   CELL_DEBUG(std::cout << "Creating cell of type " << m_type << std::endl);
}

Cell::~Cell()
{
   // delete m_type -- automatic??
}

int Cell::GetType()
{
   return m_type;
}

void Cell::SetType(int type)
{
#if 0
   // REMOVED SINCE IT HIDES THE ERROR TO THE LOG
   if (type > NUMBER_OF_TYPES)
      type = EMPTY;
   
   if (type < 0) {
      type = EMPTY;
   }
#endif
   m_type = type;
   CELL_DEBUG(std::cout << "setting type:" << GetTypeString() << std::endl);
}

std::string Cell::GetTypeString()
{
	switch (m_type) {
	   case Cell::GROW_NORTH:
	      return "1";
	      break;
	   case Cell::GROW_SOUTH:
	      return "2";
	      break;
	   case Cell::GROW_EAST:
	      return "3";
	      break;
	   case Cell::GROW_WEST:
	      return "4";
	      break;
	   case Cell::TYPE_BLACK:
	      return "5";
	      break;
	   case Cell::TYPE_WHITE:
	      return "6";
	      break;
	   case Cell::DONT_CARE:
	      return "7";
	      break;
	   case Cell::EMPTY:
	      return "8";
	      break;
	   case Cell::TYPE_RED:
	      return "9";
	      break;
	   case Cell::TYPE_BLUE:
	      return "10";
	      break;
	   case Cell::TYPE_GREEN:
	      return "11";
	      break;
	   default:
	      return "?";
	      break;
	     
	}
}


#define CELL_COLLECTION_DEBUG(x) 
//#define CELL_COLLECTION_DEBUG(x) x

CellCollection::CellCollection(Cell* center, 
	                       Cell* north, 
		               Cell* south, 
		               Cell* east, 
		               Cell* west):
   m_c(center),
   m_n(north),
   m_s(south),
   m_e(east),
   m_w(west)
{
   CELL_COLLECTION_DEBUG(std::cout << "CellCollection created:" << std::endl 
	 << "  " << m_n->GetTypeString() << std::endl
	 << m_w->GetTypeString()
	 << m_c->GetTypeString() 
	 << m_e->GetTypeString() << std::endl
	 << "  " << m_s->GetTypeString() << std::endl);
}

CellCollection::~CellCollection()
{
   // only pointers to m_board cells. not removed here. 
}

void CellCollection::AddRule(Rule* rule) 
{
   m_rules.push_back(rule);
}

/* evaluate rules for the calling cell
 * return cell's next state
 * 
 * Returns the result for a rule-match between the rule conditions and the
 * cell's neighbourhood (cellcollection). DOES NOT CHECK IF THE NEW RESULT IS
 * SIMILAR TO THE ORIGINAL CELL TYPE. USE CellCollectionRuleMatch() TO AVOID
 * THIS!
 *
 * parameter: int printMatch - for printing to std-out the matched rule
 */
int CellCollection::Evaluate(int printMatch)
{
   int result = m_c->GetType();
   bool ruleMatched = false;
   // Center (d=dont care, for hver rule sjekkes alle betingelsene 1-5)
   
   std::vector<Rule* >::iterator it;

   // check rules until one matches
   for (it = m_rules.begin(); it != m_rules.end() && !ruleMatched; ++it)
   {
      if ( (*it)->Match(m_c,m_n,m_s,m_e,m_w) ) 
      {
         ruleMatched=true; 
	 int action = (*it)->GetAction();
	 switch (action) {
	   case Cell::GROW_NORTH:
	      result = m_n->GetType();
	      break;
	   case Cell::GROW_SOUTH:
	      result = m_s->GetType();
	      break;
	   case Cell::GROW_EAST:
	      result = m_e->GetType();
	      break;
	   case Cell::GROW_WEST:
	      result = m_w->GetType();
	      break;
	   case Cell::TYPE_WHITE:
	      result = Cell::TYPE_WHITE;
	      break;
	   case Cell::TYPE_BLACK:
	      result = Cell::TYPE_BLACK;
	      break;
	   case Cell::TYPE_RED:
	      result = Cell::TYPE_RED;
	      break;
	   case Cell::TYPE_BLUE:
	      result = Cell::TYPE_BLUE;
	      break;
	   case Cell::TYPE_GREEN:
	      result = Cell::TYPE_GREEN;
	      break;
	   default:
	      break;
	 }
	 // only print out matched rules 
	 if (printMatch == 1 && ruleMatched == true)
	 {
	    std::cout << "MATCH: ";
	    (*it)->ToString();
	 }
      }
   }
   CELL_COLLECTION_DEBUG(std::cout << "Evaluated cell to " 
	                           << (new Cell(result))->GetTypeString() 
				   << std::endl);
   return result;
}

// Returns the center of a cellcollection
// Used instead of Evaluate(int) if RuleMatch() is false
int CellCollection::GetCellCollectionCenter()
{
   return m_c->GetType();
}

// returns the rule number from 0 to N, or -1 if no match
int CellCollection::GetMatchedRuleNumber()
{
   int rulenumber = -1;
   bool ruleMatched = false;
   std::vector<Rule* >::iterator it;
   for (it = m_rules.begin(); it != m_rules.end() && !ruleMatched; ++it)
   {
      rulenumber++;
      if ( (*it)->Match(m_c,m_n,m_s,m_e,m_w) ) 
      {
	 ruleMatched = true;
	 return rulenumber;
      }
   }
   // no match
   if (ruleMatched == false)
   {
      return -1;
   }
}

/* Returns true if ruleset contains match for this cellcollection
 *
 * A MATCH IS DEFINED AS A MATCH BETWEEN THE RULE CONDITIONS AND THE CELL'S
 * NEIGHBOURHOOD (CELLCOLLECTION) ONLY IF THE RESULT (VALUE/CELL TYPE) IS 
 * NOT GROWTH OF EMPTY CELLS INTO EMPTY CELLS (EMPTY-IN-EMPTY GROWTH).
 *
 * PRESERVING CHANGE RESULTS, E.G. CHANGE WHITE-TO-WHITE IS A MATCH SINCE
 * WE WANT TO NOW IF AN INDIVIDUAL HAS A STATIC OR DYNAMIC FINAL PATTERN/
 * PHENOTYPE. EVEN THOUGH THIS RESULTS IN NO DIFFERENCE IN THE PHENOTYPE.
 * (AN ACTUAL CHANGE IN THE PHENOTYPE AND CELL'S TYPE) */
bool CellCollection::CellCollectionRuleMatch()
{
   bool ruleMatched = false;
   // check rules until one matches
   std::vector<Rule* >::iterator it;
   for (it = m_rules.begin(); it != m_rules.end() && !ruleMatched; ++it)
   {
      if ( (*it)->Match(m_c,m_n,m_s,m_e,m_w) ) 
      {
	 int action = (*it)->GetAction();
	 switch (action) 
	 {
	   case Cell::GROW_NORTH:
	      // no action if we grow empty-to-empty cells
	      if (m_c->GetType() == Cell::EMPTY && 
		  m_n->GetType() == Cell::EMPTY) 
	      {
		 // ruleMatched still false 
	      }
	      else
	      {
		 ruleMatched=true; 
	      }
	      break;
	   case Cell::GROW_SOUTH:
	      // no action if we grow empty-to-empty cells
	      if (m_c->GetType() == Cell::EMPTY && 
		  m_s->GetType() == Cell::EMPTY) 
	      {
		 // ruleMatched still false 
	      }
	      else
	      {
		 ruleMatched=true; 
	      }
	      break;
	   case Cell::GROW_EAST:
	      // no action if we grow empty-to-empty cells
	      if (m_c->GetType() == Cell::EMPTY && 
		  m_e->GetType() == Cell::EMPTY) 
	      {
		 // ruleMatched still false 
	      }
	      else
	      {
		 ruleMatched=true; 
	      }
	      break;
	   case Cell::GROW_WEST:
	      // no action if we grow empty-to-empty cells
	      if (m_c->GetType() == Cell::EMPTY && 
		  m_w->GetType() == Cell::EMPTY) 
	      {
		 // ruleMatched still false 
	      }
	      else
	      {
		 ruleMatched=true; 
	      }
	      break;
	   case Cell::TYPE_WHITE:
	      // if we change white-to-white cells
	      if (m_c->GetType() == Cell::TYPE_WHITE) 
	      {
		 // report white-to-white celltype changes 
		 ruleMatched=true; 
	      }
	      else
	      {
		 ruleMatched=true; 
	      }
	      break;
	   case Cell::TYPE_BLACK:
	      // if we change black-to-black cells
	      if (m_c->GetType() == Cell::TYPE_BLACK) 
	      {
		 // report black-to-black celltype changes 
		 ruleMatched=true; 
	      }
	      else
	      {
		 ruleMatched=true; 
	      }
	      break;
	   case Cell::TYPE_RED:
	      // if we change red-to-red cells
	      if (m_c->GetType() == Cell::TYPE_RED) 
	      {
		 // report red-to-red celltype changes 
		 ruleMatched=true; 
	      }
	      else
	      {
		 ruleMatched=true; 
	      }
	      break;
	   case Cell::TYPE_BLUE:
	      // if we change blue-to-blue cells
	      if (m_c->GetType() == Cell::TYPE_BLUE) 
	      {
		 // report blue-to-blue celltype changes 
		 ruleMatched=true; 
	      }
	      else
	      {
		 ruleMatched=true; 
	      }
	      break;
	   case Cell::TYPE_GREEN:
	      // if we change green-to-green cells
	      if (m_c->GetType() == Cell::TYPE_GREEN) 
	      {
		 // report green-to-green celltype changes 
		 ruleMatched=true; 
	      }
	      else
	      {
		 ruleMatched=true; 
	      }
	      break;
	   default:
	      break;
	 }
      }
   }
   return ruleMatched;
}

#define CELL_ARRAY_DEBUG(x)
//#define CELL_ARRAY_DEBUG(x) x

CellArray::CellArray(int rows, int cols):
   m_array(0),
   m_rows(rows),
   m_cols(cols)
{
   CELL_ARRAY_DEBUG(std::cout <<  "Creating CellArray"  << std::endl);
   
   m_array = new CellGrid();
   
   for (int row = 0 ; row < m_rows ; row++) {
      CellRow* cellRow = new CellRow();
      for(int col = 0 ; col < m_cols ; col++) {
	 cellRow->push_back(new Cell());
      }
      m_array->push_back(cellRow);
   }

   CELL_ARRAY_DEBUG(std::cout <<  "CellArray created"  << std::endl);
}

CellArray::~CellArray () 
{
   CellGrid::iterator gIt;
   CellRow::iterator rIt;

   for (gIt = m_array->begin(); gIt != m_array->end(); ++gIt)
   {
      CellRow* currentRow = *gIt;
      for (rIt = currentRow->begin(); rIt != currentRow->end(); ++rIt) 
      {
	 delete *rIt; // calls ~Cell()
      }
      delete currentRow;
   }
}

Cell* CellArray::Get(int row, int col) 
{
   Cell* c = m_array->at(row)->at(col);
   CELL_ARRAY_DEBUG(std::cout << "("<< row<< ","<< col << ")"
	 << "Got cell of type: " << c->GetTypeString() << std::endl);
   return c;
}


int CellArray::GetType(int row, int col) 
{
   return Get(row,col)->GetType();
}

void CellArray::SetType(int row, int col, int type)
{
   Get(row,col)->SetType(type);
}

#define CELL_COLLECTION_DEBUG(x)
//#define CELL_COLLECTION_DEBUG(x) x
CellCollection * CellArray::CreateCellCollection(int row, int col) 
{
  CELL_COLLECTION_DEBUG(std::cout << "Creating collection for " 
	<< row << " " << col << std::endl);
  int crow=row;
  int ccol=col;

  if (crow < 0 || crow >= m_rows) {
     return 0;
  }
  if (ccol < 0 || ccol >= m_cols) {
     return 0;
  }

  int nrow = crow-1;
  if (nrow < 0) nrow=nrow+m_rows;
  int srow = crow+1;
  if (srow >= m_rows) srow=srow-m_rows;
  int wcol = ccol-1;
  if (wcol < 0) wcol=wcol+m_cols;
  int ecol = ccol+1;
  if (ecol >= m_cols) ecol=ecol-m_cols;
  
  CELL_COLLECTION_DEBUG(std::cout << "crow:" << crow << " " 
				  << "ccol:" << ccol << " "
				  << "nrow:" << nrow << " "
				  << "srow:" << srow << " "
				  << "ecol:" << ecol << " "
				  << "wcol:" << wcol << " "
				  << std::endl);

  return new CellCollection ( 
	Get(crow,ccol),
	Get(nrow,ccol),
	Get(srow,ccol),
	Get(crow,ecol),
	Get(crow,wcol)
	);
}

std::string CellArray::ToString() 
{
   std::string retVal="";
   for (int i = 0 ; i < m_rows;i++) 
   {
      for (int i = 0 ; i < m_cols;i++) 
      {
	 retVal += m_array->at(m_rows)->at(m_cols)->GetTypeString()+" ";
      }
   }
   return retVal;
}

