/*******************************************************************************
 * config.hpp
 * class representation of the configuration in exploreed.conf 
 *
 * Copyright (C) 2006 Joacim Thomassen
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or any later version. 
 * (See accompanying file LICENSE.txt or copy at 
 * http://www.gnu.org/copyleft/gpl.html)
 */

#ifndef CONFIG_HPP
#define CONFIG_HPP

#include <ostream>
#include <string>
#include <stdexcept> // runtime_error
#include <boost/regex.hpp>
#include "tinyxml/tinyxml.h"
#include <iostream>
#include <iomanip>
#include <string>

using namespace std;

// Represent the configuration file
class Config
{
   private:
      std::string m_name;
      std::string m_species;
      std::string m_evolution_method;
      std::string m_initial_pattern;
      std::string m_initial_type;
      std::string m_terminator;
      std::string m_mutation_type;
      int m_population_size, m_genome_size, m_rule_size, m_development_steps;
      int m_generations, m_phenotype_size, m_mutation_rate;
      bool m_store_genome, m_seed_population;
   public:
      Config() {}
      ~Config() {}
      Config(const std::string& name, 
	     const std::string& species,
	     const std::string& evolution_method,
	     const std::string& initial_pattern,
	     const std::string& initial_type,
	     const std::string& terminator,
	     const std::string& mutation_type,
	     const int population_size,
	     const int genome_size,
	     const int rule_size,
	     const int development_steps,
	     const int generations,
	     const int phenotype_size,
	     const int mutation_rate,
	     const bool store_genome,
	     const bool seed_population )
	 : m_name(name),
	   m_species(species),
	   m_evolution_method(evolution_method),
	   m_initial_pattern(initial_pattern),
	   m_initial_type(initial_type),
	   m_terminator(terminator),
	   m_mutation_type(mutation_type),
	   m_population_size(population_size),
	   m_genome_size(genome_size),
	   m_rule_size(rule_size),
	   m_development_steps(development_steps),
	   m_generations(generations),
	   m_phenotype_size(phenotype_size),
	   m_mutation_rate(mutation_rate),
	   m_store_genome(store_genome),
	   m_seed_population(seed_population)
      {}

      std::string name() const { return m_name; }
      std::string species() const { return m_species; }
      std::string evolution_method() const { return m_evolution_method; }
      std::string initial_pattern() const { return m_initial_pattern; }
      std::string initial_type() const { return m_initial_type; }
      std::string terminator() const { return m_terminator; }
      std::string mutation_type() const { return m_mutation_type; }
      int population_size() const { return m_population_size; }
      int genome_size() const { return m_genome_size; }
      int rule_size() const { return m_rule_size; }
      int development_steps() const { return m_development_steps; }
      int generations() const { return m_generations; }
      int phenotype_size() const { return m_phenotype_size; }
      int mutation_rate() const { return m_mutation_rate; }
      bool store_genome() const { return m_store_genome; }
      bool seed_population() const { return m_seed_population; }
      void setName(const std::string& name) { m_name = name; }
      void setSpecies(const std::string& species) { m_species = species; }
      void setInitial_pattern(const std::string& initial_pattern) 
      { 
	 m_initial_pattern = initial_pattern; 
      }
      void setInitial_type(const std::string& initial_type)
      {
	 m_initial_type = initial_type; 
      }
      void setTerminator(const std::string& terminator) 
      { 
	 m_terminator = terminator; 
      }
      void setMutation_type(const std::string& mutation_type) 
      { 
	 m_mutation_type = mutation_type; 
      }
      void setPopulation_size(const int& population_size) 
      { 
	 m_population_size = population_size; 
      }
      void setGenome_size(const int& genome_size)
      { 
	 m_genome_size = genome_size; 
      }
      void setRule_size(const int& rule_size)
      { 
	 m_rule_size = rule_size; 
      }
      void setDevelopment_steps(const int& development_steps) 
      { 
	 m_development_steps = development_steps; 
      }
      void setGenerations(const int& generations) 
      { 
	 m_generations = generations; 
      }
      void setPhenotype_size(const int& phenotype_size) 
      { 
	 m_phenotype_size = phenotype_size; 
      }
      void setEvolution_method(const std::string& evolution_method) 
      { 
	 m_evolution_method = evolution_method; 
      }
      void setStore_genome(const bool& store_genome) 
      { 
	 m_store_genome = store_genome; 
      }
      void setSeed_population(const bool& seed_population) 
      { 
	 m_seed_population = seed_population; 
      }
      void setMutation_rate(const bool& mutation_rate) 
      { 
	 m_mutation_rate = mutation_rate; 
      }
      // print configuration setup for the given exploration this
      void printConfigSetup() const
      {
	 ios_base::fmtflags flags = std::cout.flags();
	 int swidth = 20;
	 int iwidth = 4;
	 std::cout << std::endl;
	 std::cout << "=================================="
		   << "==================================" << std::endl;
	 std::cout << "#" << ". " << this->name() << std::endl;
	 std::cout << "=================================="
		   << "==================================" << std::endl;
	 std::cout << left
		   << setw(swidth) << "genome_size: "
		   << right
		   << setw(iwidth) << this->genome_size()
		   << setw(iwidth) << " "
		   << left
		   << setw(swidth) << "spicies: "
		   << left 
		   << setw(iwidth) << this->species()
		   << std::endl
		   << left
		   << setw(swidth) << "rule size: "
		   << right
		   << setw(iwidth) << this->rule_size()
		   << setw(iwidth) << " "
		   << left
		   << setw(swidth) << "terminator: "
		   << left 
		   << setw(swidth) << this->terminator()
		   << std::endl
		   << left
		   << setw(swidth) << "phenotype size: "
		   << right
		   << setw(iwidth) << this->phenotype_size()
		   << setw(iwidth) << " "
		   << left
		   << setw(swidth) << "initial pattern: "
		   << left 
		   << setw(swidth) << this->initial_pattern()
		   << std::endl
		   << left
		   << setw(swidth) << "development steps: "
		   << right
		   << setw(iwidth) << this->development_steps()
		   << setw(iwidth) << " "
		   << left
		   << setw(swidth) << "initial type: "
		   << left
		   << setw(swidth) << this->initial_type()
		   << std::endl
		   << left
		   << setw(swidth) << "population size: "
		   << right
		   << setw(iwidth) << this->population_size()
		   << setw(iwidth) << " "
		   << left
		   << setw(swidth) << "evolution method: "
		   << left
		   << setw(iwidth) << this->evolution_method()
		   << std::endl
		   << left
		   << setw(swidth) << "generations: "
		   << right
		   << setw(iwidth) << this->generations()
		   << setw(iwidth) << " "
		   << left
		   << setw(swidth) << "store genome: "
		   << left 
		   << boolalpha
		   << setw(swidth) << this->store_genome()
		   << std::endl
		   << left
		   << setw(swidth) << " "
		   << right 
		   << setw(iwidth) << " " 
		   << setw(iwidth) << " "
		   << left
		   << setw(swidth) << "seed population: "
		   << left
		   << setw(swidth) << this->seed_population()
		   << std::endl;
	 std::cout << "----------------------------------"
		   << "----------------------------------" << std::endl;
	 std::cout.flags(flags);
      }
};

// Represent the configuration file's run elements
class Run 
{
   private:
      std::string m_name;
      int m_priority;
      int m_runs;
      int m_output;
      int m_log_level;
   public:
      Run() {}
      ~Run() {}
      Run(const std::string& name, 
	  const int priority,
	  const int runs,
	  const int output,
	  const int log_level)
	 : m_name(name),
	   m_priority(priority),
	   m_runs(runs),
	   m_output(output),
	   m_log_level(log_level)
      {}
      std::string name() const { return m_name; }
      int priority() const { return m_priority; }
      int runs() const { return m_runs; }
      int output() const { return m_output; }
      int log_level() const { return m_log_level; }
      void setName(const std::string& name) { m_name = name; }
      void setPriority(const int& priority) { m_priority = priority; }
      void setRuns(const int& runs) { m_runs = runs; }
      void setOutput(const int& output) { m_output = output; }
      void setLog_level(const int& log_level) { m_log_level = log_level; }
};

#endif
