/*******************************************************************************
 * exploreed.cpp - explore evolution for (artificial) development
 * Exploreed is a tool for exploring design by evolution in an evolutionary 
 * hardware system. This was my Master in Computer Science, Spring 2006.
 *
 * Copyright (C) 2006 Joacim Thomassen
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * (See accompanying file LICENSE.txt or copy at
 * http://www.gnu.org/copyleft/gpl.html)
 *
 * Code compatibility:
 * Written for use under GNU/Linux, tested on Gentoo and Ubuntu
 * C++ ISO, Tested on GCC 3.4.4(x86) and GCC 4.0.2 (PowerPC/amd64)
 * Uses: STL
 *       Boost    (boost.org)
 *       tinyXML  (simple non-standard XML parser) 
 *
 * To contact the author visit: http://net.homelinux.org
 * Exploreed version 0.2, may 2006.
 */ 

// Use boost library to avoid the Year 2038 Bug, for date and time.
#include "boost/date_time/posix_time/posix_time.hpp"
#include "boost/date_time/gregorian/gregorian.hpp"
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>
#include <unistd.h>
#include <iostream>
#include <iomanip>
#include <string>
#include <fstream>
#include <ctime>
#include <cstdlib>
#include <vector>
#include "exploration.hpp"
#include "xmlparser.hpp"
#include "config.hpp"
#include "tools.hpp"

using namespace std;
using namespace boost::posix_time;
using namespace boost::gregorian;

// sort the runs in runList according to their priority
bool runSort(const Run& r1, const Run& r2)
{
   return r1.priority() <= r2.priority();
}

/* configuration file can be specified with option -c [filename]
 * if not, the program expects to find exploreed.conf
 * 
 * FIXME: messy main() method - move log redirecting, timekeeping, runSort(), to
 * tools.cpp 
 */
int main(int argc, char* argv[])
{
   // take total time of program execution
   ptime programStarted = ptime(second_clock::local_time());
   
   std::string configfile = "exploreed.conf";	// default configfile

   // handle program options given as parameters on the command line
   int opt;
   while ((opt = getopt(argc, argv, "c:s:")) != -1)
   {
      switch(opt)
      {
	 case 'c':
	    configfile = optarg; // alternativ configfile
	    std::cout << "Using config file: " << configfile << std::endl;
	    break;
	 case 's':
	    Tools::GetInstance()->SetSeed(optarg);
	    std::cout << "Using seed file: " << optarg << std::endl;
	    break;
	 case ':':
	    std::cout << "option needs a value" << std::endl;
	    break;
	 case '?':
	    std::cout << "unknown option" << std::endl;
	    break;
      }
   }
   for(; optind < argc; optind++)
      std::cout << "argument: " << argv[optind] << std::endl;

   // log events to name_date_time.exp - redirect every cout to file
   // open file (with buffer) for writing
   vector<string> filname_units;
   string filename;
   filname_units.push_back("summary");
   filname_units.push_back(to_iso_string(programStarted.date()));
   filname_units.push_back(to_iso_string(programStarted.time_of_day()));
   filname_units.push_back(string(".exp"));
   Tools::GetInstance()->join(filname_units, '_', filename);
   std::ofstream file(filename.c_str());
   // save output buffer of the passed stream
   std::streambuf* strm_buffer = std::cout.rdbuf();
   // redirect output to the file
   std::cout.rdbuf(file.rdbuf());

   std::cout << "Explore Evolution for Artificial Development 0.2" << std::endl;

   Xmlparser xmlpars(configfile);
   vector<Run> runList = xmlpars.ImportRun();
   vector<Config> configList = xmlpars.ImportConfig();

   std::cout << std::endl;
   std::cout << ":: Exploration runs ::" << std::endl;

   int counter = 0; // count the number of explorations
   // sort multiple runs according to priority
   if (runList.size() > 1)
   {
      // sort runList 
      std::sort(runList.begin(), runList.end(), runSort);
   }
   // execute elements in runList in order
   std::vector<Run>::iterator runIt;
   for (runIt = runList.begin(); runIt != runList.end(); ++runIt)
   {
      // FIXME: optimize connection between runList and configList
      // search configList for runIt.name
      // find configuration parameters for this run
      std::vector<Config>::iterator it;
      for (it = configList.begin(); it != configList.end(); ++it)
      {
	 boost::regex runName((*runIt).name());
	 if (boost::regex_match( (*it).name(), runName))
	 {
	    (*it).printConfigSetup();
	    counter = 0; // reset runcounter for each exploration setup
	    // repeat exploration, exploreed.conf:<runs> number of times
	    for (int j = 0; j < (*runIt).runs(); j++)
	    {
	       counter++;
	       // show and start time of exploration
	       ptime timeStarted = ptime(second_clock::local_time());
	       // start exploration
	       Exploration xplor((*it).name(), 
		                 (*runIt).log_level(), 
				 (*runIt).output(), 
				 (*it), 
				 timeStarted);
	       xplor.Start();
	       // show end time of exploration and duration
	       ptime timeEnded = ptime(second_clock::local_time());
	       time_period explorationDuration = time_period( timeStarted, 
							      timeEnded );
	       std::cout << "Run " << counter << ", duration: " 
			 << explorationDuration.length() 
			 << std::endl;
	    }
	 }
      }
   }
   
   // show total end time of program and duration
   ptime programEnded = ptime(second_clock::local_time());
   time_period programDuration = time_period(programStarted, programEnded);
   std::cout << std::endl << "Started: " << programStarted << std::endl;
   std::cout << "Ended: " << programEnded << std::endl;
   std::cout << "Total duration: " << programDuration.length() << std::endl;

   // restore old output buffer of the passed stream
   std::cout.rdbuf(strm_buffer);
   return 0;
}

