/*******************************************************************************
 * flagindividual.cpp
 * flag pattern individuals with neutral mapping between genome and rules
 *
 * Copyright (C) 2006 Joacim Thomassen
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or any later version. 
 * (See accompanying file LICENSE.txt or copy at 
 * http://www.gnu.org/copyleft/gpl.html)
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <typeinfo>
#include "math.h"
#include "flagindividual.hpp"
#include "cell.hpp"
#include "rule.hpp"
#include "tools.hpp"

#define INDIVIDUAL_DEBUG(x)
//#define INDIVIDUAL_DEBUG(x) x

// Initial constructor used in first generation
FlagIndividual::FlagIndividual( int genome_size, 
                                int phenotype_size,
				int development_steps,
				bool seed_population ) 
   : Individual(genome_size, phenotype_size, development_steps, seed_population)
{
   GenomeFactory();
   InitializeBoard();
}

// Constructor used by Clone() between generations 
FlagIndividual::FlagIndividual(Individual* i)
   : Individual(i)
{
   InitializeBoard();
}

// Creates a new individual for the Clone() function
Individual* FlagIndividual::NextIndividual(Individual* i)
{
   return new FlagIndividual(i);
}

FlagIndividual::~FlagIndividual()
{
   std::vector<Rule*>::iterator ruleIt;
   for(ruleIt = m_rules.begin(); ruleIt != m_rules.end(); ++ruleIt) 
   {
      delete (*ruleIt);
   }
   delete m_board;
}

//#define NEUTRAL_INDIVIDUAL_DEBUG(x)
#define NEUTRAL_INDIVIDUAL_DEBUG(x) x

// Initial creator of genome, used in first generation
void FlagIndividual::GenomeFactory()
{
   INDIVIDUAL_DEBUG(std::cout 
	     << "Flag GenomeFactory created Individual @" << this << std::endl
	     << "Type: " << typeid(*this).name() 
	     << ", Genome_size: " << m_genome_size << std::endl);
   NEUTRAL_INDIVIDUAL_DEBUG(std::cout << "I@" << this << ": ");

   /* check if genome should be created from a given seed or be randomly
    * initialized which is the default initializationi */
   if (m_seeded)
   {
      /* copy genome from seed */
      m_genome = Tools::GetInstance()->GetSeed();
      std::vector<int>::iterator ggene;
      for (ggene = m_genome.begin(); ggene != m_genome.end(); ++ggene)
      {
	 std::cout << (*ggene) << " ";
      }
      std::cout << std::endl;
   }
   else
   {
      /* make random genome for individual */
      for (int i = 0 ; i < m_genome_size ; i++ ) 
      {
	 int gene = rand()%m_genome_size;
	 m_genome.push_back(gene);
	 NEUTRAL_INDIVIDUAL_DEBUG(std::cout << gene << " ");
      }
   }
   NEUTRAL_INDIVIDUAL_DEBUG(std::cout << std::endl);
}

// Map genom into rules
void FlagIndividual::RuleFactory()
{
   INDIVIDUAL_DEBUG(std::cout << typeid(*this).name() 
	                      << " entering NEUTRAL RuleFactory:" 
			      << std::endl);
   // flush previous rules in individual
   // Remember to delete old rules
   m_rules.clear();
   
   // produce new rules from genome
   int ruleNum = 0;
   int genomeLength = m_genome.size();
   int numRules = genomeLength/6; // FIXME: ruleSize is a parameter in conf.
   
   for (ruleNum = 0; ruleNum < numRules ; ruleNum ++) 
   {
      m_rules.push_back(new Rule(GetAction(ruleNum), 
                                 GetCenter(ruleNum),
				 GetNorth(ruleNum) ,
				 GetSouth(ruleNum) ,
				 GetEast(ruleNum)  ,
				 GetWest(ruleNum)   ));
   }
}

#define DEVELOP_DEBUG(x)
//#define DEVELOP_DEBUG(x) x

/* Develop individual from genotype to phenotype. 
 * 
 * For each development_step this method is repeated, until final 
 * development_step is reached, and the fully grown phenotype is completed. 
 * Then the evaluation in Evaluate() determines the fitness of this 
 * individual's phenotype.
 */
void FlagIndividual::Develop()
{
   // Copy the board for next development stage
   CellArray* nextBoard = new CellArray(m_rows, m_cols);

   /* check my neighbours against my rules
    * UPDATES ALL CELLS INTO NEXTBOARD
    * either to its previous type or to the new one
    */
   for (int i = 0; i < m_rows; i++)
   {
      for (int j = 0; j < m_cols; j++) 
      {
	 // my neighbours
	 CellCollection* c = m_board->CreateCellCollection(i,j);
	 // cycle my rules for this cell(i,j)'s neighbours (CellCollection) 
	 std::vector<Rule*>::iterator ruleIt;
	 for(ruleIt = m_rules.begin() ; ruleIt != m_rules.end() ; ++ruleIt) {
	    c->AddRule(*ruleIt);
	 }
         nextBoard->SetType(i,j,c->Evaluate(0));
	 // Free memory used by temporal CellCollection
	 delete c; // use auto_ptr to avoid leaks on exceptions
      }
   }
   // Free memory used by the current board CellArray
   delete m_board; // m_board pointer ->NULL !!!
   // the new board state after a completed development step
   m_board = nextBoard; 
   DEVELOP_DEBUG(std::cout << "Board after development step:\n" 
			   << this->ToString()  << std::endl);
}

#define EVALUATE_DEBUG(x)
//#define EVALUATE_DEBUG(x) x

/* Evaluate each individual/phenotype
 *
 * Generates rules from the genome and evaluates them to give a fitness rating.
 * If the individual is capable of development, it's allowed to develop. After
 * its final development step the phenotype is fully developed and can be 
 * evaulated for fitness rating of the individual.
 */
void FlagIndividual::Evaluate()
{
   // check if i have already been evaluated
   if (m_fitness < 0.0)
   {
      // illegal until my rules prove otherwise
      bool anyLegalRules = false;
      // mark myself as evaluated
      this->m_fitness = 0.1;

      // Generate rules from genome
      this->RuleFactory();

      // allow individuals with at least one legal rule to develop
      std::vector<Rule*>::iterator ruleIt;
      for(ruleIt = m_rules.begin() ; ruleIt != m_rules.end() ; ++ruleIt)
      {
	 if((*ruleIt)->IsLegal())
	 {
	    // increase fitness by each legal rule
	    m_fitness += 20.0;
	    anyLegalRules = true;
	    
	    INDIVIDUAL_DEBUG(std::cout << "Legal rule: ");
	 }
	 m_fitness += (*ruleIt)->RuleFitness();
      }
   
      /* ONLY RUN DEVELOPMENT ON INDIVIDUALS WITH AT LEAST ONE LEGAL RULE
       * (only legal rules will work)
       */
      if (anyLegalRules)
      {
	 INDIVIDUAL_DEBUG(std::cout << "Developing step: ");
	 // Let the Individual grow/develop into its final phenotype 
	 for (int i = 0; i < GetDevelopment_steps(); i++) 
	 {
	    EVALUATE_DEBUG(std::cout << "Develop Individual @" << this 
		                     << " step:" << i << "\n" 
				     << this->ToString() << std::endl);
	    INDIVIDUAL_DEBUG(std::cout << i << " ");
	    Develop(); 
	 }
	 INDIVIDUAL_DEBUG(std::cout << std::endl);
	 bool black = false;
	 bool white = false;
	 bool red = false;
	 double blacks = 1;
	 double whites = 1;
	 double reds = 1;
	 double blues = 1;
	 double greens = 1;
	 double row_blacks = 0;
	 double row_whites = 0;
	 double row_reds = 0;
	 double row_blues = 0;
	 double row_greens = 0;

	 double proliferation_fitness = 0;
	 double cellscore = 5.0;

	 int phenotype_size = GetCols() * GetRows();
	 std::vector<double> row_distribution;
	 std::vector<int> rowtypes;

	 // Evaluate the full grown phenotype
	 
	 /* PROLIFERATION FITNESS ASSIGNMENT ---------------------------------*/

	 for ( int row = 0 ; row < m_rows ; row++ ) 
	 {
	   for ( int col = 0 ; col < m_cols ; col++ ) 
	   {
	      // travers phenotype, evaluate and credit proliferation
	      if (m_board->GetType(row, col) == Cell::TYPE_WHITE) 
	      {
		 m_fitness += cellscore;
		 m_phenotype_fitness += cellscore;
		 proliferation_fitness += cellscore;
		 white = true;
		 whites++;
		 row_whites++;
	      }
	      else if (m_board->GetType(row, col) == Cell::TYPE_BLACK) 
	      {
		 m_fitness += cellscore;
		 m_phenotype_fitness += cellscore;
		 proliferation_fitness += cellscore;
		 white = true;
		 blacks++;
		 row_blacks++;
	      }
	      else if (m_board->GetType(row, col) == Cell::TYPE_BLUE) 
	      {
		 m_fitness += cellscore;
		 m_phenotype_fitness += cellscore;
		 proliferation_fitness += cellscore;
		 white = true;
		 blues++;
		 row_blues++;
	      }
	      else if (m_board->GetType(row, col) == Cell::TYPE_GREEN)
	      {
		 m_fitness+=cellscore;
		 m_phenotype_fitness+=cellscore;
		 proliferation_fitness += cellscore;
		 black = true;
		 greens++;
		 row_greens++;
	      }
	      else if (m_board->GetType(row, col) == Cell::TYPE_RED)
	      {
		 m_fitness+=cellscore;
		 m_phenotype_fitness+=cellscore;
		 proliferation_fitness += cellscore;
		 red = true;
		 reds++;
		 row_reds++;
	      }
	   }
	   if ((row_reds >= row_blues) && (row_reds >= row_greens) &&
	       (row_reds > 0)) 
	   {
	      row_distribution.push_back(row_reds);
	      rowtypes.push_back(Cell::TYPE_RED);
	   }
	   else 
	   {
	      if ((row_blues >= row_greens) && (row_blues >= row_reds) &&
	          (row_blues > 0)) 
	      {
		 row_distribution.push_back(row_blues);
		 rowtypes.push_back(Cell::TYPE_BLUE);
	      }
	      else
	      {
		 if ((row_greens >= row_reds) && (row_greens >= row_blues) &&
	             (row_greens > 0)) 
		 {
		    row_distribution.push_back(row_greens);
		    rowtypes.push_back(Cell::TYPE_GREEN);
		 }
		 // if row is empty
		 else
		 {
		    rowtypes.push_back(Cell::EMPTY);
		 }
	      }
	   }
	   // zero out row counts before next
	   row_reds = 0;
	   row_whites = 0;
	   row_blacks = 0;
	   row_greens = 0;
	   row_blues = 0;
	 }
	 std::cout << "whites: " << whites-1 << std::endl
	           << "blacks: " << blacks-1 << std::endl
	           << "blues: " << blues-1 << std::endl
	           << "greens: " << greens-1 << std::endl
	           << "reds: " << reds-1 << std::endl;

	 std::cout << "proliferation fitness: " 
		   << proliferation_fitness 
		   << std::endl;
	 /* END OF PROLIFERATION FITNESS ASSIGNMENT --------------------------*/


	 // keep the possibility to only have 2 celltypes for testing purposes
	 if (Cell::NUMBER_OF_TYPES == 8)
	 {
	    /* RATIO FITNESS ASSIGNMENT FOR 2 CELLTYPES ----------------------*/

	    // credit balanace of types
	    double target_ratio = 3.0/7.0;
	    if (blacks > whites)
	    {
	       double ratio = whites/blacks;
	       std::cout << "ratio: " << ratio 
		         << " target_ratio: " << target_ratio
			 << " abs(target_r - r) = " 
			 << fabs(target_ratio - ratio) 
			 << std::endl;
	       m_fitness += ((1.0 - fabs(target_ratio - ratio)) * 
		            ((25.0 * (whites + blacks))));
	       m_phenotype_fitness += ((1.0 - fabs(target_ratio - ratio)) * 
		                      ((25.0 * (whites + blacks))));
	       std::cout << "balancefitness given: " 
		         << m_phenotype_fitness 
			 << std::endl;
	    }
	    if (whites >= blacks)
	    {
	       double ratio = blacks/whites;
	       std::cout << "ratio: " << ratio 
		         << " target_ratio: " << target_ratio
			 << " abs(target_r - r) = " 
			 << fabs(target_ratio - ratio) 
			 << std::endl;
	       m_fitness += ((1.0 - fabs(target_ratio - ratio)) * 
		            ((25.0 * (whites + blacks))));
	       m_phenotype_fitness += ((1.0 - fabs(target_ratio - ratio)) * 
		                      ((25.0 * (whites + blacks))));
	       std::cout << "balancefitness given: " 
		         << m_phenotype_fitness 
			 << std::endl;
	    }
	    std::cout << "New Fitness of Individual@" << this << ": " 
		      << m_fitness << " phenotype fitness: "
		      << m_phenotype_fitness << std::endl;
	    /* END OF RATIO FITNESS ASSIGNMENT FOR 2 CELLTYPES ---------------*/
	 }
	 

	 /* 3 visible/phenotypic celltypes (and 0 or 2 not expressed types)
	  *
	  * For the computations the 3 visible celltypes are now assigned to
	  * the variables min, med and max. This assignment is dynamic and 
	  * might change for each evaluation. It is solely determined by the 
	  * number of visible cells of each celltype in the phenotype, the 
	  * type_distribution. */
	 if (Cell::NUMBER_OF_TYPES > 8)
	 {
	    /* RATIO FITNESS ASSIGNMENT FOR 3 CELLTYPES ----------------------*/

	    // celltype distribution, given in % 0.0 - 1.0 (0% - 100%)
	    double target_ratio_min = 0.3; // always keep this one smallest
	    double target_ratio_med = 0.3; 
	    double target_ratio_max = 0.4; // always keep this one largest
	    // actual number of each celltype in phenotype
	    double min; 
	    double med; 
	    double max; 

	    // only visible celltypes are evaluated for fitness score
	    // put celltypes that are visible in type_distribution
	    std::vector<double> type_distribution;
	    type_distribution.push_back(blues);
	    type_distribution.push_back(greens);
	    type_distribution.push_back(reds);
	    sort(type_distribution.begin(), type_distribution.end());
	    min = type_distribution.at(0); // always kept smallest
	    med = type_distribution.at(1);
	    max = type_distribution.at(2); // always kept largest

	    // number_of_visible_cell can be used in Cell distribution fitness
	    // if fitness score needs to get higher/balanced
	    // NOT USED FOR NOW, gives no higher Success Ratio at the moment 
	    // 8.may 2006
	    double number_of_visible_cells = 0.0; 
	    std::vector<double>::iterator tdit;
	    for (tdit = type_distribution.begin();
		 tdit != type_distribution.end(); ++tdit)
	    {
	       number_of_visible_cells += (*tdit);
	    }

	    // CELL RATIO (RELATIVE NUMBER OF EACH CELLTYPE TO PHENOTYPE SIZE)

	    double ratio_min = min/phenotype_size;
	    double ratio_med = med/phenotype_size;
	    double ratio_max = min/phenotype_size;

	    // want a deviation between [0-1], for the fitness calculations
	    double n_min = target_ratio_min;
	    if (target_ratio_min < 0.5)
	    {
	       n_min = fabs(target_ratio_min - 1.0);
	    }
	    double n_med = target_ratio_med;
	    if (target_ratio_med < 0.5)
	    {
	       n_med = fabs(target_ratio_med - 1.0);
	    }
	    double n_max = target_ratio_max;
	    if (target_ratio_max < 0.5)
	    {
	       n_max = fabs(target_ratio_max - 1.0);
	    }
	    double deviation_min = (fabs(target_ratio_min - ratio_min)) / n_min;
	    double deviation_med = (fabs(target_ratio_med - ratio_med)) / n_med;
	    double deviation_max = (fabs(target_ratio_max - ratio_max)) / n_max;

	    // TYPE RATIO (BETWEEN CELLTYPES)
	    
	    double number_of_visible_types = type_distribution.size();

	    double target_ratio_a = target_ratio_min / target_ratio_max;
	    double target_ratio_b = target_ratio_med / 
	                            (target_ratio_min + target_ratio_max);

	    double ratio_a = min/max;
	    double ratio_b = med/(min + max);

	    // want a deviation between [0-1], for the fitness calculations
	    double n_a = target_ratio_a;
	    if (target_ratio_a < 0.5)
	    {
	       n_a = fabs(target_ratio_a - 1.0);
	    }
	    double n_b = target_ratio_b;
	    if (target_ratio_b < 0.5)
	    {
	       n_b = fabs(target_ratio_b - 1.0);
	    }
	    double deviation_a = (fabs(target_ratio_a - ratio_a)) / n_a;
	    double deviation_b = (fabs(target_ratio_b - ratio_b)) / n_b;
	    

	    // Cell distribution fitness
	    double fitness_min = (1200.0 * (1.0 - deviation_min));
	    double fitness_med = (1200.0 * (1.0 - deviation_med));
	    double fitness_max = (1200.0 * (1.0 - deviation_max));

	    // Celltype-ratio fitness
	    double fitness_a = (1.0 - deviation_a) * 
	                       ((fitness_min + fitness_max) / 2.0);
	    double fitness_b = (1.0 - deviation_b) * fitness_med; 
	    
	    // FITNESS TIPPING POINT
	    // ASSIGN RATIO FIRST AFTER CELLGROWTH HAS EXPANDED TO MAXIMUM
	    // PHENOTYPE SIZE (phenotypefitness > 50*phenotype_size)
	    if (m_phenotype_fitness > ((cellscore*phenotype_size)-1))
	    {
	       m_fitness += (fitness_a + fitness_b);
	       m_phenotype_fitness += (fitness_a + fitness_b);
	    }

	    std::cout << "ratio_a: " << ratio_a 
	              << " target_ratio_a: " << target_ratio_a
	              << " ratio_b: " << ratio_b 
		      << " target_ratio_b: " << target_ratio_b
		      << std::endl;

	    std::cout << "ratiofitness: " << (fitness_a + fitness_b) 
	              << std::endl;
	    /* END OF RATIO FITNESS ASSIGNMENT FOR 3 CELLTYPES ---------------*/
	    
	    /* FLAG FITNESS ASSIGNMENT ---------------------------------------*/
	    // Evaluate the phenotype pattern distribution
	    
#if 0
	    /* This fitness was too restrictive, not used for now.*/
	    // contains celltype(5,6,8,9) for homogene rows or 0 non-homogene
	    // credit each fully homogene row
	    int identical_sum;
	    for ( int row = 0 ; row < m_rows ; row++ ) 
	    {
	       int current_celltype;
	       identical_sum = 0;
	       for ( int col = 0 ; col < m_cols ; col++ ) 
	       {
		  // only credit non EMPTY rows
		  if (m_board->GetType(row, col) != Cell::EMPTY)
		     {
		     if (col == 0) 
		     {
			current_celltype = m_board->GetType(row, col);
			identical_sum++;
		     }
		     if ((col > 0) && 
			 (m_board->GetType(row, col) == current_celltype))
		     {
			identical_sum++;
		     }
		  }
	       }
	       // assign fitness if an entire row is of identical celltype
	       if (identical_sum == GetCols())
	       {
		  // homogene row
		  rowtypes.push_back(current_celltype);
		  m_fitness += ((fitness_a + fitness_b)) / GetRows();
		  m_phenotype_fitness += ((fitness_a + fitness_b)) / 
		                         GetRows();
	       }
	       
	       else
	       {
		  // not a homogene row
		  rowtypes.push_back(0);
	       }
	    }
#endif
	    // IDENTICAL_SUM FITNESS, given for the number of identical 
	    // celltypes in each row 
	    double total_identical_fitness = 0;
	    std::vector<double>::iterator rsi;
	    for (rsi = row_distribution.begin(); 
		 rsi != row_distribution.end(); ++rsi)
	    {
	       total_identical_fitness += 10.0 * (*rsi); 
	    }
	    m_fitness += total_identical_fitness;
	    m_phenotype_fitness += total_identical_fitness;
	    std::cout << "Identical fitness: " << total_identical_fitness
		      << std::endl;
	    
	    // max_identical_fitness used in connection fitness
	    double max_identical_fitness = (pow(GetRows(),2.0)*10.0);

	    // CONNECTION FITNESS, given for connections between rows with
	    // equal celltype dominans
	    int current_rowtype = 0;
	    int connections = 0;
	    double total_connection_fitness = 0;
	    std::vector<int>::iterator rti;
	    for (rti = rowtypes.begin(); rti != rowtypes.end(); ++rti)
	    {
	       // compare rows with equal celltype dominans
	       if ((*rti) != Cell::EMPTY)
	       {
		  /* a more numerous sequence of connections will give better 
		   * fitness than breaked sequences since first row in a 
		   * connectionsequence do not increase fitness */
		  if (connections == 0) 
		  {
		     current_rowtype = (*rti);
		     connections = 1;
		  }
		  else
		  {
		     /* check for equal-row connection. only give increased
		      * fitness until a constricted sequence length, given by
		      * (GetRows()*target_ratio_med) */
		     if (((*rti) == current_rowtype) && (connections > 0)
			   && (connections < (GetRows()*target_ratio_med)))
		     {
			connections++;
			/* only credit connections after the identical_fitness
			 * has reached to half of its maximum fitness, e.g. 
			 * each row contains a high number of its dominant 
			 * celltype */
			if (total_identical_fitness > max_identical_fitness/2.0)
			{
			   m_fitness += (fitness_min +
				         fitness_med +
					 fitness_max) / 10.0;
			   m_phenotype_fitness += (fitness_min +
				                   fitness_med +
						   fitness_max) / 10.0;
			}
			// used for logging purposes
			total_connection_fitness += (fitness_min + 
			                             fitness_med +
						     fitness_max) / 10.0;
		     }
		     else
		     {
			// current rowtype was not equal to the previous
			connections = 1;
			current_rowtype = (*rti);
		     }
		  }
	       }
	       // EMPTY celltype dominans are not credited
	       else
	       {
		  connections = 0;
		  current_rowtype = 0;
	       }
	    }
	    std::cout << "Connection fitness: " << total_connection_fitness 
	              << std::endl;
	    /* END OF FLAG FITNESS ASSIGNMENT --------------------------------*/

	 }
	 std::cout << "New Fitness of Individual@" << this << ": " 
	           << m_fitness << " phenotype fitness: "
		   << m_phenotype_fitness << std::endl;
      }
   }
}

void FlagIndividual::Mutate()
{
   int mutationPoint, newGene, nmbrOfMPoints; 
   nmbrOfMPoints = rand()%3+1;
   INDIVIDUAL_DEBUG(std::cout << "Mutating genome" << std::endl);
   for (int i = 0; i < nmbrOfMPoints; i++)
   {
      mutationPoint = rand()%m_genome.size();
      newGene = rand()%m_genome.size();
      INDIVIDUAL_DEBUG(std::cout << " gene: " << mutationPoint 
	                         << " into: " << newGene);
      this->m_genome.at(mutationPoint) = newGene;
   }
   INDIVIDUAL_DEBUG(std::cout << std::endl);
}

// must construct a new one, since board has to be reset to initial state
Individual* FlagIndividual::Clone()
{
   return NextIndividual(this);
}

// create empty board with startup conditions
void FlagIndividual::InitializeBoard()
{
   m_board = new CellArray(m_rows, m_cols);
   /* FIXME: random placement? not sure if it matters as edges wraps around 
    *        SetType(rand()%m_rows,rand()%m_cols,rand()%8+1)
    *        don't matter but should instead be placed in center for better view
    *        of the early development, and to comply to Gunnar Tufte's work.
    */
   m_board->SetType(0%m_rows,1%m_cols,Cell::TYPE_GREEN);
}
