/**********************************************************************
 * individual.cpp - individual class for each individual in population
 * contains all common individual functions
 *
 * Copyright (C) 2006 Joacim Thomassen
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or any later version. 
 * (See accompanying file LICENSE.txt or copy at 
 * http://www.gnu.org/copyleft/gpl.html)
 */

/* for srand() and rand(), use GNU Scientific lib. for better
 * and faster number generator when optimizing */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <algorithm>
#include "individual.hpp"
#include "cell.hpp"
#include "rule.hpp"

/* protected constructors, only used to set parameters */

Individual::Individual(int genome_size,
	               int phenotype_size, 
	               int development_steps,
		       bool seed_population) 
   : m_genome(),
     m_genome_size(genome_size),
     m_rows(phenotype_size),  
     m_cols(phenotype_size),
     m_development_steps(development_steps),
     m_seeded(seed_population),
     m_fitness(-1.0),
     m_phenotype_fitness(-1.0),
     m_board(0)
{} 

Individual::Individual(Individual* i) 
   : m_genome(i->GetGenome()),
     m_genome_size(i->GetGenomeSize()),
     m_rows(i->GetRows()),  
     m_cols(i->GetRows()),
     m_development_steps(i->GetDevelopment_steps()),
     m_seeded(i->GetSeed_state()),
     m_fitness(-1.0),
     m_phenotype_fitness(-1.0),
     m_board(0)
{} 

/* public functions */

int Individual::GetAction(int ruleNumber)
{
   int action_adress;
   action_adress = m_genome[ruleNumber*6+0];
   return ((action_adress + m_genome[action_adress])/2)%Cell::NUMBER_OF_TYPES+1;
}
int Individual::GetCenter(int ruleNumber)
{
   int action_adress;
   action_adress = m_genome[ruleNumber*6+1];
   return ((action_adress + m_genome[action_adress])/2)%Cell::NUMBER_OF_TYPES+1;
}
int Individual::GetNorth(int ruleNumber)
{
   int action_adress;
   action_adress = m_genome[ruleNumber*6+2];
   return ((action_adress + m_genome[action_adress])/2)%Cell::NUMBER_OF_TYPES+1;
}
int Individual::GetSouth(int ruleNumber)
{
   int action_adress;
   action_adress = m_genome[ruleNumber*6+3];
   return ((action_adress + m_genome[action_adress])/2)%Cell::NUMBER_OF_TYPES+1;
}
int Individual::GetEast(int ruleNumber)
{
   int action_adress;
   action_adress = m_genome[ruleNumber*6+4];
   return ((action_adress + m_genome[action_adress])/2)%Cell::NUMBER_OF_TYPES+1;
}
int Individual::GetWest(int ruleNumber)
{
   int action_adress;
   action_adress = m_genome[ruleNumber*6+5];
   return ((action_adress + m_genome[action_adress])/2)%Cell::NUMBER_OF_TYPES+1;
}

void Individual::PrintRuleset()
{
   std::cout << "My ruleset: " << std::endl;

   std::vector<Rule*>::iterator ruleIt;
   for(ruleIt = m_rules.begin(); ruleIt != m_rules.end(); ++ruleIt) 
   {
      (*ruleIt)->ToString();
   }
}

void Individual::PrintGenome()
{
   std::cout << "My genome: ";

   std::vector<int>::iterator geneIt;
   for (geneIt = m_genome.begin(); geneIt != m_genome.end(); ++geneIt) 
   {
      std::cout << (*geneIt) << " ";
   }
   std::cout << std::endl;
}

std::string Individual::PrintPhenotype()
{
   //FIXME: Check if GetTypeString() leaks
   std::string phenotypeString;
   for (int i = 0 ; i < m_rows ; i++) {
     for (int j = 0 ; j < m_cols ; j++) {
	phenotypeString+=m_board->Get(i,j)->GetTypeString()+" ";
     }
     phenotypeString+="\n";
   }
   return phenotypeString;
}

std::string Individual::ToString()
{
   char * tmp = new char[1024];
   sprintf( tmp,
	    "Individual@0x%08x with fitness %.2f, phenotypefitness %.2f\n",
	    this,
	    m_fitness,
	    m_phenotype_fitness);

   std::string objString = tmp;
   for (int i = 0 ; i < m_rows ; i++) {
     for (int j = 0 ; j < m_cols ; j++) {
	objString+=m_board->Get(i,j)->GetTypeString()+" ";
     }
     objString+="\n";
   }
   delete[] tmp; // or auto_ptr to avoid leak on exceptions
   return objString;
}

// Prints out development steps for an individual 
void Individual::ToString_Development()
{
   std::cout << "-------------------------------------------------------\n"
             << " Development steps of final Individual @" << this << "\n"
	     << "-------------------------------------------------------\n"
	     << std::endl;
   this->PrintGenome();
   // generate ruleset 
   this->RuleFactory();
   // must mark legal rules, or they are not used during development
   std::vector<Rule*>::iterator ruleIt;
   for(ruleIt = m_rules.begin() ; ruleIt != m_rules.end() ; ++ruleIt)
   {
      (*ruleIt)->IsLegal();
   }
   this->PrintRuleset();
   //FIXME: run development if no legal rules? must inform user if no output.
   for (int s = 0; s < GetDevelopment_steps(); s++) 
   {
      int matched_rules = 0;
      std::vector<int> matched_rulenumbers;
      
      // Copy the board for next development stage
      CellArray* nextBoard = new CellArray(m_rows, m_cols);

      // check my neighbours against my rules
      for (int i = 0; i < m_rows; i++)
      {
	 for (int j = 0; j < m_cols; j++) 
	 {
	    // my neighbours
	    CellCollection* c = m_board->CreateCellCollection(i,j);
	    // cycle my rules   
	    std::vector<Rule*>::iterator ruleIt;
	    for(ruleIt = m_rules.begin() ; ruleIt != m_rules.end() ; ++ruleIt) 
	    {
	       c->AddRule(*ruleIt);
	    }
	    // do Evaluation of MATCHED rule sets
	    if (c->CellCollectionRuleMatch())
	    {
	       matched_rules++;
	       std::cout << "Cell(" << i << "," << j << ") " ;
	       nextBoard->SetType(i,j,c->Evaluate(1));
	       matched_rulenumbers.push_back(c->GetMatchedRuleNumber());
	    }
	    else
	    {
	       nextBoard->SetType(i,j,c->GetCellCollectionCenter());
	    }
	    // Free memory used by temporal CellCollection
	    delete c; // use auto_ptr to avoid leaks on exceptions
	 }
      }
      // Free memory used by the current board CellArray
      delete m_board; // m_board pointer ->NULL !!!
      m_board = nextBoard;
      std::cout << "Development_step: " << s << std::endl;

      // Evaluate each development step
      int blacks = 0;
      int whites = 0;
      int reds = 0;
      int blues = 0;
      int greens = 0;
      for ( int row = 0 ; row < m_rows ; row++ ) {
	for ( int col = 0 ; col < m_cols ; col++ ) {
	   // travers phenotype, evaluate and register proliferation
	   if (m_board->GetType(row, col) == Cell::TYPE_WHITE) 
	   {
	      whites++;
	   }
	   else if (m_board->GetType(row, col) == Cell::TYPE_BLACK)
	   {
	      blacks++;
	   }
	   else if (m_board->GetType(row, col) == Cell::TYPE_RED)
	   {
	      reds++;
	   }
	   else if (m_board->GetType(row, col) == Cell::TYPE_BLUE)
	   {
	      blues++;
	   }
	   else if (m_board->GetType(row, col) == Cell::TYPE_GREEN)
	   {
	      greens++;
	   }
	}
      }
      std::cout << "Black_types: " << blacks << std::endl;
      std::cout << "White_types: " << whites << std::endl;
      std::cout << "Red_types: " << reds << std::endl;
      std::cout << "Blue_types: " << blues << std::endl;
      std::cout << "Green_types: " << greens << std::endl;
      std::cout << "Matched_rules: " << matched_rules << std::endl;
      // sort rulenumbers
      sort(matched_rulenumbers.begin(), matched_rulenumbers.end());
      // remove duplicates, for plotting if/not rule trigged at dev.step
      std::vector<int>::iterator pos;
      pos = unique(matched_rulenumbers.begin(), matched_rulenumbers.end());
      matched_rulenumbers.erase(pos, matched_rulenumbers.end());
      std::cout << "Triggered_rules: ";
      for (pos = matched_rulenumbers.begin(); 
	   pos != matched_rulenumbers.end(); 
	   ++pos)
      {
	 std::cout << (*pos) << " ";
      }
      std::cout << std::endl;
	 
      std::cout << "Printing types:" << std::endl;
      std::cout << PrintPhenotype() << std::endl;

   }
}
