/*******************************************************************************
 * neutralindividual.cpp
 * eg. chessboardindividual with *-1 mapping between genome and rules,
 * also known as neutral mapping with neutral network
 *
 * Copyright (C) 2006 Joacim Thomassen
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or any later version. 
 * (See accompanying file LICENSE.txt or copy at 
 * http://www.gnu.org/copyleft/gpl.html)
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <typeinfo>
#include "neutralindividual.hpp"
#include "individual.hpp"
#include "cell.hpp"
#include "rule.hpp"
#include "tools.hpp"
/* for srand() and rand(), use GNU Scientific lib. for better
 * and faster number generator when optimizing */


#define INDIVIDUAL_DEBUG(x)
//#define INDIVIDUAL_DEBUG(x) x

// Initial constructor used in first generation
NeutralIndividual::NeutralIndividual( int genome_size, 
                                      int phenotype_size,
				      int development_steps,
				      bool seed_population ) 
   : Individual(genome_size, phenotype_size, development_steps, seed_population)
{
   GenomeFactory();
   InitializeBoard();
}

// Constructor used by Clone() between generations 
NeutralIndividual::NeutralIndividual(Individual* i)
   : Individual(i)
{
   InitializeBoard();
}

// Creates a new individual for the Clone() function
Individual* NeutralIndividual::NextIndividual(Individual* i)
{
   return new NeutralIndividual(i);
}

NeutralIndividual::~NeutralIndividual()
{
   std::vector<Rule*>::iterator ruleIt;
   for(ruleIt = m_rules.begin(); ruleIt != m_rules.end(); ++ruleIt) 
   {
      delete (*ruleIt);
   }
   delete m_board;
}

//#define NEUTRAL_INDIVIDUAL_DEBUG(x)
#define NEUTRAL_INDIVIDUAL_DEBUG(x) x

// Initial creator of genome, used in first generation
void NeutralIndividual::GenomeFactory()
{
   INDIVIDUAL_DEBUG(std::cout 
	     << "Neutral GenomeFactory created Individual @" << this 
	     << std::endl
	     << "Type: " << typeid(*this).name() 
	     << ", Genome_size: " << m_genome_size << std::endl);
   NEUTRAL_INDIVIDUAL_DEBUG(std::cout << "I@" << this << ": ");

   /* check if genome should be created from a given seed or be randomly
    * initialized which is the default initializationi */
   if (m_seeded)
   {
      /* copy genome from seed */
      m_genome = Tools::GetInstance()->GetSeed();
      std::vector<int>::iterator ggene;
      for (ggene = m_genome.begin(); ggene != m_genome.end(); ++ggene)
      {
	 std::cout << (*ggene) << " ";
      }
      std::cout << std::endl;
   }
   else
   {
      /* make random genome for individual */
      for (int i = 0 ; i < m_genome_size ; i++ ) 
      {
	 int gene = rand()%m_genome_size;
	 m_genome.push_back(gene);
	 NEUTRAL_INDIVIDUAL_DEBUG(std::cout << gene << " ");
      }
   }
   NEUTRAL_INDIVIDUAL_DEBUG(std::cout << std::endl);
}

// Map genom into rules
void NeutralIndividual::RuleFactory()
{
   INDIVIDUAL_DEBUG(std::cout << typeid(*this).name() 
	                      << " entering NEUTRAL RuleFactory:" 
			      << std::endl);
   // flush previous rules in individual
   // Remember to delete old rules
   m_rules.clear();
   
   // produce new rules from genome
   int ruleNum = 0;
   int genomeLength = m_genome.size();
   int numRules = genomeLength/6; // FIXME: ruleSize is a parameter in conf.
   
   for (ruleNum = 0; ruleNum < numRules ; ruleNum ++) 
   {
      m_rules.push_back(new Rule(GetAction(ruleNum), 
                                 GetCenter(ruleNum),
				 GetNorth(ruleNum) ,
				 GetSouth(ruleNum) ,
				 GetEast(ruleNum)  ,
				 GetWest(ruleNum)   ));
   }
}

#define DEVELOP_DEBUG(x)
//#define DEVELOP_DEBUG(x) x

/* Develop individual from genotype to phenotype. 
 * 
 * For each development_step this method is repeated, until final 
 * development_step is reached, and the fully grown phenotype is completed. 
 * Then the evaluation in Evaluate() determines the fitness of this 
 * individual's phenotype.
 */
void NeutralIndividual::Develop()
{
   // Copy the board for next development stage
   CellArray* nextBoard = new CellArray(m_rows, m_cols);

   /* check my neighbours against my rules
    * UPDATES ALL CELLS INTO NEXTBOARD
    * either to its previous type or to the new one
    */
   for (int i = 0; i < m_rows; i++)
   {
      for (int j = 0; j < m_cols; j++) 
      {
	 // my neighbours
	 CellCollection* c = m_board->CreateCellCollection(i,j);
	 // cycle my rules for this cell(i,j)'s neighbours (CellCollection) 
	 std::vector<Rule*>::iterator ruleIt;
	 for(ruleIt = m_rules.begin() ; ruleIt != m_rules.end() ; ++ruleIt) {
	    c->AddRule(*ruleIt);
	 }
         nextBoard->SetType(i,j,c->Evaluate(0));
	 // Free memory used by temporal CellCollection
	 delete c; // use auto_ptr to avoid leaks on exceptions
      }
   }
   // Free memory used by the current board CellArray
   delete m_board; // m_board pointer ->NULL !!!
   // the new board state after a completed development step
   m_board = nextBoard;
   DEVELOP_DEBUG(std::cout << "Board after development step:\n" 
			   << this->ToString()  << std::endl);
}

#define EVALUATE_DEBUG(x)
//#define EVALUATE_DEBUG(x) x
void NeutralIndividual::Evaluate()
{
   // check if i have already been evaluated
   if (m_fitness < 0.0)
   {
      // illegal until my rules prove otherwise
      bool anyLegalRules = false;
      // mark myself as evaluated
      this->m_fitness = 0.1;

      // Generate rules from genome
      this->RuleFactory();

      // allow individuals with at least one legal rule to develop
      std::vector<Rule*>::iterator ruleIt;
      for(ruleIt = m_rules.begin() ; ruleIt != m_rules.end() ; ++ruleIt)
      {
	 if((*ruleIt)->IsLegal())
	 {
	    // increase fitness by each legal rule
	    m_fitness += 20.0;
	    anyLegalRules = true;
	    
	    INDIVIDUAL_DEBUG(std::cout << "Legal rule: ");
	    // FIXME: BUG r141 #1. (*ruleIt)->ToString();
	 }
	 /* OLD: Suppress illegal rules, (DON'T decrement m_fitness)
	 else 
	 { 
	    //m_fitness -= 10;
	    //legal = false;
	 }*/
	 
	 /* Run detailed Fitness feedback on rules, converge too slow.
	  * Stimulates the evolution of legal rules 
	  */
	 m_fitness += (*ruleIt)->RuleFitness();
      }
   
      /* ONLY RUN DEVELOPMENT ON INDIVIDUALS WITH AT LEAST ONE LEGAL RULE
       * (only legal rules will work)
       */
      if (anyLegalRules)
      {
	 INDIVIDUAL_DEBUG(std::cout << "Developing step: ");
	 // Let the Individual grow/develop into its final phenotype 
	 for (int i = 0; i < GetDevelopment_steps(); i++) 
	 {
	    EVALUATE_DEBUG(std::cout << "Develop Individual @" << this 
		                     << " step:" << i << "\n" 
				     << this->ToString() << std::endl);
	    INDIVIDUAL_DEBUG(std::cout << i << " ");
	    Develop(); 
	 }
	 INDIVIDUAL_DEBUG(std::cout << std::endl);
	 
	 // Evaluate the full grown phenotype
	 for ( int i = 0 ; i < m_rows ; i++ ) {
	   for ( int j = 0 ; j < m_cols ; j++ ) {
	      if ( (i+j)%2 == 0) {
		if (m_board->GetType(i,j) == Cell::TYPE_WHITE) {
		   m_fitness+=50.0;
		   m_phenotype_fitness+=50.0;
		}
	      }
	      else {
		if (m_board->GetType(i,j) == Cell::TYPE_BLACK) {
		   m_fitness+=50.0;
		   m_phenotype_fitness+=50.0;
		}
	      }
	   }
	 }
      }
      std::cout << "New Fitness of Individual@" << this 
	                       << ": " << m_fitness << " phenotype fitness: "
			       << m_phenotype_fitness << std::endl;
   }
}
#if 0 
// OLD MUTATION() FUNCTION
//Individual* NeutralIndividual::Mutate()
void NeutralIndividual::Mutate()
{
#if 0
   std::cout << this << " of type " << typeid(*this).name() 
	     << " entering NEUTRAL Mutate:" << std::endl;
#endif
   int mutationPoint, newGene, nmbrOfMPoints; //, thresholdFitness;
   /* mutationrate changes to fitness score
    * possible number of mutation points 1-23
    */
#if 0
   thresholdFitness = 70; // DON'T SET HERE!
   if (m_fitness < thresholdFitness)
   {
      nmbrOfMPoints = rand()%6+1; // not 0 points
   }
   else
   {
      nmbrOfMPoints = rand()%3+1;
   }
#endif
   nmbrOfMPoints = rand()%3+1;
   //nmbrOfMPoints = rand()%23+1;
   for (int i = 0; i < nmbrOfMPoints; i++)
   {
      mutationPoint = rand()%m_genome.size();
      newGene = rand()%m_genome.size();
      m_genome.at(mutationPoint) = newGene;
   }
   //return NextIndividual(this);
}
#endif

void NeutralIndividual::Mutate()
{
   int mutationPoint, newGene, nmbrOfMPoints; 
   nmbrOfMPoints = rand()%3+1;
   INDIVIDUAL_DEBUG(std::cout << "Mutating genome" << std::endl);
   for (int i = 0; i < nmbrOfMPoints; i++)
   {
      mutationPoint = rand()%m_genome.size();
      newGene = rand()%m_genome.size();
      INDIVIDUAL_DEBUG(std::cout << " gene: " << mutationPoint 
	                         << " into: " << newGene);
      this->m_genome.at(mutationPoint) = newGene;
   }
   INDIVIDUAL_DEBUG(std::cout << std::endl);
}

// must construct a new one, since board has to be reset to initial state
Individual* NeutralIndividual::Clone()
{
#if 0
   std::cout << this << " of type " << typeid(*this).name() 
	     << " entering universal Clone:" << std::endl;
#endif
   return NextIndividual(this);
}

// create empty board with initial zygote (start cell) 
void NeutralIndividual::InitializeBoard()
{
   m_board = new CellArray(m_rows, m_cols);
   m_board->SetType(0%m_rows,1%m_cols,Cell::TYPE_BLACK);
}

