/*******************************************************************************
 * population.cpp
 * During the exploration one population object is used to contain all the 
 * individuals. During the evolution AddIndividual is used to populate the 
 * m_individuals vector of the next population for each new generation.
 *
 * Copyright (C) 2006 Joacim Thomassen
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or any later version. 
 * (See accompanying file LICENSE.txt or copy at 
 * http://www.gnu.org/copyleft/gpl.html)
 */
#include <iostream>
#include <algorithm>
#include "population.hpp"
#include <boost/regex.hpp>
#include <string>

bool genomeSort(Individual* i1, Individual* i2)
{
   return i1->GetFitness() >= i2->GetFitness();
}

// Creates the specified population
Population::Population(int population_size,
		       int genome_size,
		       int phenotype_size, // always AxA, symmetry!
		       int development_steps,
		       std::string species,
		       bool seed_population)
   : m_individuals(),
     m_population_size(population_size),
     fittest(0)
{
   boost::regex n_pattern("neutral");
   boost::regex s_pattern("simple");
   boost::regex f_pattern("flag");
   // flag, *-1 mapped genotype->rules
   if (boost::regex_match(species, f_pattern)) 
   {
      for (int i = 0; i < m_population_size; i++)
      {
	 AddIndividual(new FlagIndividual( genome_size, 
		                           phenotype_size,
					   development_steps,
					   seed_population ));
      }
   }
   else 
   {
      // neutral, *-1 mapped genotype->rules
      if (boost::regex_match(species, n_pattern)) 
      {
	 for (int i = 0; i < m_population_size; i++)
	 {
	    AddIndividual(new NeutralIndividual( genome_size, 
						 phenotype_size,
						 development_steps,
	                                         seed_population ));
	 }
      }
      else
      {
	 // simple, 1-1 mapped genotype->rules
	 if (boost::regex_match(species, s_pattern)) 
	 {
	    for (int i = 0; i < m_population_size; i++)
	    {
	       AddIndividual(new SimpleIndividual( genome_size, 
						   phenotype_size,
						   development_steps, 
	                                           seed_population ));
	    }
	 }
	 else
	 {
	    std::cout << "Sorry, no such evolutionary computation available"
		      << std::endl;
	 }
      }
   }
}

Population::~Population()
{
   std::vector<Individual*>::iterator indIt;
   for(indIt = m_individuals.begin(); indIt != m_individuals.end(); ++indIt)
   {
      delete (*indIt);
   }
   m_individuals.clear();
}

// Add the given individual to population
void Population::AddIndividual(Individual* i)
{
   if (i != 0)
   {
      m_individuals.push_back(i);
   }
}

#define POPULATION_DEBUG(x)
//#define POPULATION_DEBUG(x) x
// Evolution on each generation
void Population::Evolve()
{
   std::cout << "Size of population: "<< m_individuals.size() << std::endl;

   /* Evaluate and sort individuals, high fitness = high ranking (0=highest) */
   std::vector<Individual*>::iterator it;
   for (it = m_individuals.begin(); it != m_individuals.end(); ++it)
   {
      (*it)->Evaluate(); 
   }
   std::sort(m_individuals.begin(), m_individuals.end(), genomeSort);

   std::cout << "Fittest " << m_individuals.front()->ToString() 
             << std::endl;
   m_individuals.front()->PrintRuleset();
 
   /* Do Selection on population */
   Selection();

   /* Free up memory occupied by dead individuals */
   /* clear all elements from this population and free memory */
   for (it = m_individuals.begin(); it != m_individuals.end(); ++it)
   {
      delete (*it);
   }
   m_individuals.clear();
   /* Take population to next generation */
   BreedNextGeneration();
}

// Clone() and Mutate() rest of population from fittest individual
void Population::BreedNextGeneration()
{
   // ELITISM & JULIANS "MUTATIONFILL"
   /* add the best individual to the next generation's population */
   AddIndividual(fittest); 
   /* fill the remaining spaces with mutants of the best individual
      (static sized populations) */
   for (int i = 0; i < (m_population_size-1); i++) 
   {
      Individual* mutant = fittest->Clone();
      mutant->Mutate();
      AddIndividual(mutant);
      POPULATION_DEBUG(std::cout << fittest << " cloned & mutated into " 
	                         << mutant << std::endl);
   }
} 

// Survivor selection used between each generation
void Population::Selection()
{
   // ELITISM, keep only the ONE best individual
   /* store best individual and size of population for next generation */
   POPULATION_DEBUG(std::cout << m_individuals.front() << " cloned into "); 
   fittest = m_individuals.front()->Clone();
   POPULATION_DEBUG(std::cout << fittest << std::endl);
   // OTHER TYPES OF SELECTION, must use a m_selected vector of size = popsize
}

void Population::PrintPopulation()
{
   std::cout << "Population consists of: " << std::endl;
   std::vector<Individual*>::iterator pit;
   for (pit = m_individuals.begin(); pit != m_individuals.end(); ++pit)
   {
      std::cout << (*pit)->ToString() << std::endl;
   }
}

/* run development and evaluate this individual again for printout */
void Population::printDevelopment()
{
   if (fittest != 0)
      fittest->ToString_Development();
}
