/*******************************************************************************
 * xmlparser.cpp
 * simple parser using tinyxml to parse the content of exploreed.conf 
 *
 * Copyright (C) 2006 Joacim Thomassen
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or any later version. 
 * (See accompanying file LICENSE.txt or copy at 
 * http://www.gnu.org/copyleft/gpl.html)
 */

#include "tinyxml/tinyxml.h"
#include "xmlparser.hpp"
#include "config.hpp"
#include <iostream>
#include <ostream>
#include <vector>
#include <exception>
#include <stdexcept>
#include <cstdlib>
#include <cstring>
#include <string>
#include <boost/lexical_cast.hpp>

using namespace std;

Xmlparser::Xmlparser(const std::string& config_filename)
   : m_configfile(config_filename)
{
   // ready to parse
}

std::vector<Run> Xmlparser::ImportRun()
{
   try 
   {
      std::vector<Run> runList;
      bool run_found = false;

      // Parse "exploreed.conf" or userdefined configure file
      TiXmlDocument doc(m_configfile);
      if (!doc.LoadFile())
      {
	 throw runtime_error("bad parse");
      }

      // Verify that root is an config element
      TiXmlElement* root = doc.RootElement();
      if (strcmp(root->Value(), "config") != 0)
      {
	 throw runtime_error(string("bad root: ") + root->Value());
      }

      // Traverse children of root, populating the explorationList
      for ( TiXmlElement* exploration = root->FirstChildElement();
	    exploration;
	    exploration = exploration->NextSiblingElement() )
      {
	 // Verify that run corresponds to a "run" element
	 if (strcmp(exploration->Value(), "run") == 0)
	 {
	    run_found = true;
	    runList.push_back(nodeToRun(exploration));
	 }
	 else
	 {
	    // go to next element
	 }
      }
      if (run_found)
      {
	 // run element found in config file 
	 return runList;
      }
      else
      {
	 throw runtime_error( string("bad config: no run found! ") );
      }
   }
   catch (const exception& e)
   {
      std::cout << e.what() << "\n";
   }
}

std::vector<Config> Xmlparser::ImportConfig()
{
   try 
   {
      std::vector<Config> explorationList;
      bool exploration_found = false;

      // Parse "exploreed.conf" or userdefined configure file
      TiXmlDocument doc(m_configfile);
      if (!doc.LoadFile())
      {
	 throw runtime_error("bad parse");
      }

      // Verify that root is an config element
      TiXmlElement* root = doc.RootElement();
      if (strcmp(root->Value(), "config") != 0)
      {
	 throw runtime_error(string("bad root: ") + root->Value());
      }

      // Traverse children of root, populating the explorationList
      for ( TiXmlElement* exploration = root->FirstChildElement();
	    exploration;
	    exploration = exploration->NextSiblingElement() )
      {
	 // Verify that exploration corresponds to a "exploration" element
	 if (strcmp(exploration->Value(), "exploration") == 0)
	 {
	    exploration_found = true;
	    explorationList.push_back(nodeToExploration(exploration));
	 }
	 else
	 {
	    // go to next element
	 }
      }
      if (exploration_found)
      {
	 // exploration element found in config file 
	 return explorationList;
      }
      else
      {
	 throw runtime_error( string("bad config: no exploration found! ") );
      }
   }
   catch (const exception& e)
   {
      std::cout << e.what() << "\n";
   }
}

// FIXME: if setningen nedover er dårlig forma for integerløsningene
// sette value eller sjekke value !!!???

// Extract the content of an XML element that contains only text
const char* textValue(TiXmlElement* e)
{
   TiXmlNode* first = e->FirstChild();
   if ( first != 0 &&
	first == e->LastChild() &&
	first->Type() == TiXmlNode::TEXT )
   {
      // the element e has a single child, of type TEXT;
      // return the child's
      return first->Value();
   }
   else
   {
      throw runtime_error(string("bad") + e->Value() + " element");
   }
}

// Construct a Config from an "exploration" element
Config Xmlparser::nodeToExploration(TiXmlElement* exploration)
{
   Config result; // Return value
   TiXmlElement* element = exploration->FirstChildElement();

   // Read name
   if (element && strcmp(element->Value(), "name") == 0)
   {
      // The first child element of config is a "name"
      // element; use its text value to set the name of result
      result.setName(textValue(element));
   }
   else
   {
      throw runtime_error("no name attribute");
   }
   // Read genome_size 
   element = element->NextSiblingElement();
   const char* value;
   const char* value0;
   if ( element && strcmp(element->Value(), "genome_size") == 0 && 
	(value = element->Attribute("value")) )
   {
      // Cast string genome_size to it's integer value
      const int intValue = boost::lexical_cast<int>(value);
      result.setGenome_size(intValue);
   }
   else
   {
      throw runtime_error("no genom_size attribute");
   }
   // Read rule_size 
   element = element->NextSiblingElement();
   value = 0;
   if ( element && strcmp(element->Value(), "rule_size") == 0 && 
	(value = element->Attribute("value")) )
   {
      const int intValue = boost::lexical_cast<int>(value);
      result.setRule_size(intValue);
   }
   else
   {
      throw runtime_error("no rule_size attribute");
   }
   // Read species
   element = element->NextSiblingElement();
   if ( element && strcmp(element->Value(), "species") == 0 ) 
   {
      result.setSpecies(textValue(element));
   }
   else
   {
      throw runtime_error("no species attribute");
   }
   // Read phenotype_size
   element = element->NextSiblingElement();
   value = 0;
   if ( element && strcmp(element->Value(), "phenotype_size") == 0 && 
	(value = element->Attribute("value")) )
   {
      const int intValue = boost::lexical_cast<int>(value);
      result.setPhenotype_size(intValue);
   }
   else
   {
      throw runtime_error("no phenotype_size attribute");
   }
   // Read development_steps 
   element = element->NextSiblingElement();
   value = 0;
   if ( element && strcmp(element->Value(), "development_steps") == 0 && 
	(value = element->Attribute("value")) )
   {
      const int intValue = boost::lexical_cast<int>(value);
      result.setDevelopment_steps(intValue);
   }
   else
   {
      throw runtime_error("no development_steps attribute");
   }
   // Read population_size
   element = element->NextSiblingElement();
   value = 0;
   if ( element && strcmp(element->Value(), "population_size") == 0 && 
	(value = element->Attribute("value")) )
   {
      const int intValue = boost::lexical_cast<int>(value);
      result.setPopulation_size(intValue);
   }
   else
   {
      throw runtime_error("no population_size attribute");
   }
   // Read generations
   element = element->NextSiblingElement();
   value = 0;
   if ( element && strcmp(element->Value(), "generations") == 0 && 
	(element->Attribute("value")) )
   {
      value = element->Attribute("value");
      const int intValue = boost::lexical_cast<int>(value);
      result.setGenerations(intValue);
   }
   else
   {
      throw runtime_error("no generations attribute");
   }
   // Read terminator
   element = element->NextSiblingElement();
   if ( element && strcmp(element->Value(), "terminator") == 0 ) 
   {
      result.setTerminator(textValue(element));
   }
   else
   {
      throw runtime_error("no terminator attribute");
   }
   // Read initial_pattern
   element = element->NextSiblingElement();
   if ( element && strcmp(element->Value(), "initial_pattern") == 0 ) 
   {
      result.setInitial_pattern(textValue(element));
   }
   else
   {
      throw runtime_error("no initial_pattern attribute");
   }
   // Read initial_type
   element = element->NextSiblingElement();
   if ( element && strcmp(element->Value(), "initial_type") == 0 ) 
   {
      result.setInitial_type(textValue(element));
   }
   else
   {
      throw runtime_error("no initial_type attribute");
   }
   // Read evolution_method
   element = element->NextSiblingElement();
   value = 0;
   value0 = 0;
   if ( element && strcmp(element->Value(), "evolution_method") == 0 && 
	(value = element->Attribute("mutation_type")) &&
	(value0 = element->Attribute("mutation_rate")) )
   {
      const int intValue = boost::lexical_cast<int>(value0);
      result.setMutation_type(value);
      result.setMutation_rate(intValue);
   }
   else
   {
      throw runtime_error("no evolution_method attribute");
   }
   // Read store_genome
   element = element->NextSiblingElement();
   bool bool_value = 0;
   if ( element && strcmp(element->Value(), "store_genome") == 0 ) 
   {
      if (strcmp(textValue(element), "yes") == 0)
      {
	 result.setStore_genome(true);
      }
      else if (strcmp(textValue(element), "no") == 0)
      {
	 result.setStore_genome(false);
      }
      else
      {
	 throw runtime_error("wrong formed store_genome attribute");
      }
   }
   else
   {
      throw runtime_error("no store_genome attribute");
   }
   // Read seed_population
   element = element->NextSiblingElement();
   if ( element && strcmp(element->Value(), "seed_population") == 0 ) 
   {
      if (strcmp(textValue(element), "yes") == 0)
      {
	 result.setSeed_population(true);
      }
      else if (strcmp(textValue(element), "no") == 0)
      {
	 result.setSeed_population(false);
      }
      else
      {
	 throw runtime_error("wrong formed seed_population attribute");
      }
   }
   else
   {
      throw runtime_error("no seed_population attribute");
   }

   // Check that there are no more children
   element = element->NextSiblingElement();
   if (element != 0)
   {
      throw runtime_error( std::string("unexpected element:") + 
	                   element->Value());
   }
   return result;
}

// Construct a Run from an "run" element
Run Xmlparser::nodeToRun(TiXmlElement* exploration)
{
   Run result; // Return value
   TiXmlElement* element = exploration->FirstChildElement();

   // Read name
   if (element && strcmp(element->Value(), "exploration_name") == 0)
   {
      // The first child element of config is a "exploration_name"
      // element; use its text value to set the name of result
      result.setName(textValue(element));
   }
   else
   {
      throw runtime_error("no exploration_name attribute");
   }
   // Read priority 
   element = element->NextSiblingElement();
   const char* value;
   if ( element && strcmp(element->Value(), "priority") == 0 && 
	(value = element->Attribute("value")) )
   {
      // Cast string priority to it's integer value
      const int intValue = boost::lexical_cast<int>(value);
      result.setPriority(intValue);
   }
   else
   {
      throw runtime_error("no priority attribute");
   }
   // Read runs 
   element = element->NextSiblingElement();
   value = 0;
   if ( element && strcmp(element->Value(), "runs") == 0 && 
	(value = element->Attribute("value")) )
   {
      const int intValue = boost::lexical_cast<int>(value);
      result.setRuns(intValue);
   }
   else
   {
      throw runtime_error("no runs attribute");
   }
   // Read output 
   element = element->NextSiblingElement();
   value = 0;
   if ( element && strcmp(element->Value(), "output") == 0 && 
	(value = element->Attribute("value")) )
   {
      const int intValue = boost::lexical_cast<int>(value);
      result.setOutput(intValue);
   }
   else
   {
      throw runtime_error("no output attribute");
   }
   // Read log_level 
   element = element->NextSiblingElement();
   value = 0;
   if ( element && strcmp(element->Value(), "log_level") == 0 && 
	(value = element->Attribute("value")) )
   {
      const int intValue = boost::lexical_cast<int>(value);
      result.setLog_level(intValue);
   }
   else
   {
      throw runtime_error("no log_level attribute");
   }
   // Check that there are no more children
   element = element->NextSiblingElement();
   if (element != 0)
   {
      throw runtime_error( std::string("unexpected element:") + 
	                   element->Value());
   }
   return result;
}
